/*
 *	Canvace's JavaScript Game Engine, canvace.js
 *	Copyright (C) 2013  Canvace Srl  <http://www.canvace.com/>
 *
 *	Dual licensed under the MIT and GPLv3 licenses.
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Helper class that eases cross-browser mouse input management.
 *
 * @class Canvace.Mouse
 * @constructor
 * @param element {HTMLElement} The HTML element that captures mouse input. This
 * is usually the HTML5 canvas.
 */
Canvace.Mouse = function (element) {
	var moveHandlers = new Canvace.MultiSet();
	var downHandlers = new Canvace.MultiSet();
	var upHandlers = new Canvace.MultiSet();
	var dragHandlers = new Canvace.MultiSet();
	var wheelHandlers = new Canvace.MultiSet();

	var dragging = false, button = 0;
	var x0, y0;

	function wrapEventHandler(handler) {
		return function (event) {
			var rect = element.getBoundingClientRect();
			handler.call(
				element,
				event,
				event.clientX - rect.left,
				event.clientY - rect.top
				);
		};
	}

	element.addEventListener('mousedown', wrapEventHandler(function (event, x, y) {
		button = event.button;
		dragging = true;

		x0 = x;
		y0 = y;
		downHandlers.fastForEach(function (handler) {
			handler(x, y, event.button);
		});
	}), false);
	element.addEventListener('mousemove', wrapEventHandler(function (event, x, y) {
		moveHandlers.fastForEach(function (handler) {
			handler(x, y);
		});
		if (dragging) {
			dragHandlers.fastForEach(function (handler) {
				handler(x0, y0, x, y, button);
			});
			x0 = x;
			y0 = y;
		}
	}), false);
	element.addEventListener('mouseup', wrapEventHandler(function (event, x, y) {
		dragging = false;
		upHandlers.fastForEach(function (handler) {
			handler(x, y, event.button);
		});
	}), false);

	if (typeof element.onwheel !== 'undefined') {
		element.addEventListener('wheel', wrapEventHandler(function (event, x, y) {
			wheelHandlers.fastForEach(function (handler) {
				handler(x, y, -event.deltaX, -event.deltaY);
			});
		}), false);
	} else if (typeof element.onmousewheel !== 'undefined') {
		element.addEventListener('mousewheel', wrapEventHandler(function (event, x, y) {
			wheelHandlers.fastForEach(function (handler) {
				handler(x, y, event.wheelDeltaX, event.wheelDeltaY);
			});
		}), false);
	}

	// TODO capture touch events

	/**
	 * Registers the specified user-defined event handler that is invoked every
	 * time the mouse is moved over the HTML element.
	 *
	 * The specified function receives two arguments, `x` and `y`, indicating
	 * the coordinates of the mouse pointer relative to the element's left top
	 * corner. The return value is ignored.
	 *
	 * This method returns a function that unregisters the registered handler.
	 *
	 * Multiple handlers may be registered. The same handler may also be
	 * registered more than once, in which case it gets called as many times as
	 * it was registered every time the mouse is moved.
	 *
	 * If a handler is registered more than once, the function returned by this
	 * method only removes its own registration.
	 *
	 * @method onMove
	 * @param handler {Function} A user-defined handler function. The specified
	 * function receives two arguments (the coordinates of the mouse pointer)
	 * and its return value is ignored.
	 * @return {Function} A function that unregisters the registered handler.
	 * The returned function does not receive any arguments and does not return
	 * anything.
	 */
	this.onMove = function (handler) {
		return moveHandlers.add(handler);
	};

	/**
	 * Registers the specified user-defined event handler that is invoked every
	 * time a mouse button is pressed over the HTML element.
	 *
	 * The specified function receives three arguments, `x` and `y`, indicating
	 * the coordinates of the mouse pointer relative to the element's left top
	 * corner, and `button`, indicating which mouse button has been pressed
	 * (`0` for left button, `1` for middle button, `2` for right button).
	 * The return value is ignored.
	 *
	 * This method returns a function that unregisters the registered handler.
	 *
	 * Multiple handlers may be registered. The same handler may also be
	 * registered more than once, in which case it gets called as many times as
	 * it was registered every time a mouse button is pressed.
	 *
	 * If a handler is registered more than once, the function returned by this
	 * method only removes its own registration.
	 *
	 * @method onDown
	 * @param handler {Function} A user-defined handler function. The specified
	 * function receives two arguments (the coordinates of the mouse pointer)
	 * and its return value is ignored.
	 * @return {Function} A function that unregisters the registered handler.
	 * The returned function does not receive any arguments and does not return
	 * anything.
	 */
	this.onDown = function (handler) {
		return downHandlers.add(handler);
	};

	/**
	 * Registers the specified user-defined event handler that is invoked every
	 * time a mouse button is depressed over the HTML element.
	 *
	 * The specified function receives three arguments, `x` and `y`, indicating
	 * the coordinates of the mouse pointer relative to the element's left top
	 * corner, and `button`, indicating which mouse button has been pressed
	 * (`0` for left button, `1` for middle button, `2` for right button).
	 * The return value is ignored.
	 *
	 * This method returns a function that unregisters the registered handler.
	 *
	 * Multiple handlers may be registered. The same handler may also be
	 * registered more than once, in which case it gets called as many times as
	 * it was registered every time a mouse button is depressed.
	 *
	 * If a handler is registered more than once, the function returned by this
	 * method only removes its own registration.
	 *
	 * @method onUp
	 * @param handler {Function} A user-defined handler function. The specified
	 * function receives two arguments (the coordinates of the mouse pointer)
	 * and its return value is ignored.
	 * @return {Function} A function that unregisters the registered handler.
	 * The returned function does not receive any arguments and does not return
	 * anything.
	 */
	this.onUp = function (handler) {
		return upHandlers.add(handler);
	};

	/**
	 * Registers the specified user-defined event handler that is invoked every
	 * time the mouse is dragged over the HTML element.
	 *
	 * The specified function receives five arguments, `x0` and `y0`, indicating
	 * the last recorded coordinates of the mouse pointer before the drag event
	 * relative to the element's left top corner, `x` and `y`, indicating the
	 * current coordinates of the mouse pointer relative to the element's left
	 * top corner, and `button`, indicating which mouse button has initiated the
	 * drag event (`0` for left button, `1` for middle button, `2` for right
	 * button). The return value is ignored.
	 *
	 * This method returns a function that unregisters the registered handler.
	 * Multiple handlers may be registered.
	 *
	 * The same handler may also be registered more than once, in which case it
	 * gets called as many times as it was registered every time the mouse is
	 * dragged.
	 *
	 * If a handler is registered more than once, the function returned by this
	 * method only removes its own registration.
	 *
	 * @method onDrag
	 * @param handler {Function} A user-defined handler function. The specified
	 * function receives two arguments (the coordinates of the mouse pointer)
	 * and its return value is ignored.
	 * @return {Function} A function that unregisters the registered handler.
	 * The returned function does not receive any arguments and does not return
	 * anything.
	 */
	this.onDrag = function (handler) {
		return dragHandlers.add(handler);
	};

	/**
	 * Registers the specified user-defined event handler that is invoked every
	 * time the mouse wheel is used to scroll over the HTML element.
	 *
	 * The specified function receives four arguments, `x`, `y`, `dX` and `dY`,
	 * indicating the coordinates of the mouse pointer relative to the element's
	 * left top corner and the relative wheel movement in the X and Y
	 * directions. The return value is ignored.
	 *
	 * This method returns a function that unregisters the registered handler.
	 *
	 * Multiple handlers may be registered. The same handler may also be
	 * registered more than once, in which case it gets called as many times as
	 * it was registered every time the mouse is moved.
	 *
	 * If a handler is registered more than once, the function returned by this
	 * method only removes its own registration.
	 *
	 * @method onWheel
	 * @param handler {Function} A user-defined handler function. The specified
	 * function receives three arguments (the `x` and `y` coordinates of the
	 * mouse pointer, and the relative `dX` and `dY` movement of the mouse
	 * wheel) and its return value is ignored.
	 * The `dX` and `dY` arguments respectively assume positive values in case
	 * of a rightwards or upwards movement, and negative values in case of a
	 * leftwards or downwards movement.
	 * @return {Function} A function that unregisters the registered handler.
	 * The returned function does not receive any arguments and does not return
	 * anything.
	 */
	this.onWheel = function (handler) {
		return wheelHandlers.add(handler);
	};
};
